
/* ase3.c -- utilities and interface functions for the ase3
	code in PIC12F1822 devices. */

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <linux/i2c-dev.h>
#include <utilib.h>
#include "timdat.h"
#include "ase3.h"

int showchar(int cc)
{
	if(cc < 0x20 || cc >= 0x7f) return('.'); 
	return (cc); 
}

/* The 61 Hz rate isn't quite 61 Hz, makes the measuring interval
	not quite 1 second. Real rate is ftick().  */ 

double ftick()
{
	return (8.0E6 / 131072.0);
}

double secondscount(uint32_t taa)
{
	return (double) taa / ftick();
}


/* Read and calculate the scale-factor, which is stored as
	a power of 10 (2s complement) byte in EEPROM at address 10. 
	This indicates how many decades we have to shift the accumuator
	value to get the original units.  Example, for a flowmeter with
	716 pulses per liter, the scale-factor is 100, as what is being
	accumulater is really centiliters, 7.16 pulses per centiliter that 
	is.
*/
int ascale(int fd, int ba, double *pascale)
{
	int s; 
	int8_t sexp;

	*pascale = 0.0;

	s = ioctl(fd, I2C_SLAVE, ba);  /* 010 0000 */ 
	if(s == -1) return(-1); 

	/* Determine the scale-factor, which is stored as
		a power of 10 (2s complement) byte in EEPROM at address 10. */

	delay(20);
	s = i2c_smbus_write_byte_data(fd, EEPROMADDRSET_1, 10);
	if(s == -1) return(-1); 
	delay(20);
	s = i2c_smbus_write_byte(fd, EEPROMREAD_0);
	if(s == -1) return(-1); 
	delay(20);

	s = i2c_smbus_read_byte_data(fd, I2C_TX_EEPROMBYTE);
	if(s == -1) return(-1); 
	sexp = s;  /* This will sign-extend */ 
	*pascale = pow(10.0, (double) sexp); 

	// printf("%s(%d): Found scale exp %d scale %G\n", __FILE__, __LINE__, sexp, *pascale); 

	return(0); 
}

int aunits(int fd, int ba, char *punit, size_t pumax)
{
	int s; 

	punit[0] = 0;
	if(pumax < 2) return(-1);

	s = ioctl(fd, I2C_SLAVE, ba);  /* 010 0000 */ 
	if(s == -1) return(-1); 

	delay(20);
	s = i2c_smbus_write_byte_data(fd, EEPROMADDRSET_1, 11);
	if(s == -1) return(-1); 
	delay(20);
	s = i2c_smbus_write_byte(fd, EEPROMREAD_0);
	if(s == -1) return(-1); 
	delay(20);

	s = i2c_smbus_read_byte_data(fd, I2C_TX_EEPROMBYTE);
	if(s == -1) return(-1); 
	punit[0] = s;
	punit[1] = 0;

	// printf("%s(%d): Found scale exp %d scale %G\n", __FILE__, __LINE__, sexp, *pascale); 
	return(0); 
}

/* Return accumulator value and time of accumulation */ 
int avalue(int fd, int ba, double *avalue, uint32_t *ptaa)
{
	int s; 
	uint32_t raa; 
	uint32_t taa; 
	double scale10;

	s = ioctl(fd, I2C_SLAVE, ba);  /* 010 0000 */ 
	if(s == -1) return(-1); 

	s = i2c_smbus_write_byte(fd, ASNAPSHOT_0);
	if(s == -1) return(-1); 

	delay(50);
	s = read(fd, &raa, 4); 
	if(s == -1) return(-1); 
	
	delay(50);
	s = i2c_smbus_write_byte(fd, I2C_TX_SECONDS0);
	if(s == -1) return(-1); 
	delay(50);
	s = read(fd, &taa, 4); 
	if(s == -1) return(-1); 

	s = ascale(fd, ba, &scale10);
	if(s == -1) return(-1); 

	/* printf("%s(%d): Read V=%u at T=%u\n", __FILE__, __LINE__, raa, taa); */ 

	*avalue = (double) raa / (256.0 * scale10);

	*ptaa = taa;
	return(taa);
}

/* Convert any read tick-count to standard clock time. 
	avtime_x = 0 will be the time of last ticks-counter 
	reset.  */

time_t convertatime(int fd, int ba, 
	uint32_t avtime_x, 
	int *year, int *month, int *day, 
	int *hour, int *min, int *sec)
{
	struct tm meas_tm;
	time_t meast_sa;

	meast_sa = tickunixtime(fd, ba, avtime_x, NULL);

	gmtime_r(&meast_sa, &meas_tm); 
	choptm(&meas_tm, year, month, day, hour, min, sec);
	
	return(meast_sa); 
}


double uptime(int fd, int ba)
{
	int s; 
	double uptime_s;
	uint32_t devtks_x;

	s = ioctl(fd, I2C_SLAVE, ba);  /* 010 0000 */ 
	if(s == -1) return(-1); 

	/* Then get the time from the device */
	s = i2c_smbus_write_byte(fd, TSNAPSHOT_0); 
	delay(100);
	// fprintf(stderr, "%s(%d): about to read...\n", __FILE__, __LINE__);
	s = read(fd, &devtks_x, sizeof(devtks_x) );
	// fprintf(stderr, "%s(%d): read %d...\n", __FILE__, __LINE__, s);

	uptime_s = (double) (devtks_x) / ftick();

	// fprintf(stderr, "%s(%d): read %G...\n", __FILE__, __LINE__, uptime_s);

	return(uptime_s); 
}


/* Convert the tick-count to standard UNIX time. 
	avtime_x = 0 will be the time of last ticks-counter reset.  */

time_t tickunixtime(int fd, int ba, uint32_t avtime_x, double *puptime)
{
	double upts;
	time_t  nowux_sa,  meast_sa;
	time_t  uptime_s; 

	upts = uptime(fd, ba);
	time(&nowux_sa);

	// fprintf(stderr, "%s(%d): read %G...\n", __FILE__, __LINE__, upts);

	if(puptime) *puptime = upts;

	/* Do the time differnce since start first. */ 
	uptime_s = floor( upts - ( (double) avtime_x / ftick() ) + 0.5);
	meast_sa = nowux_sa - uptime_s; 
	return(meast_sa); 
}


/* Measure the frequency-counter value */
double fmeas(int fd, int ba)
{
	int s; 
	char liste[5];
	int aa;
	double fa;

	s = ioctl(fd, I2C_SLAVE, ba); 
	if(s == -1) 
	{
		// fprintf(stderr, "Bus address %d ", ba); 
		// perror("I2C_SLAVE");
		return(-1.0);
	}

	s = i2c_smbus_write_byte(fd, I2C_TX_TMR1L); 
	if(s < 0)
	{
		// fprintf(stderr, "smbus_write_byte I2C_TX_TMR1L "); 
		// perror(i2cdev); 
		return(-1.0);
	}

	delay(50);

	s = read(fd, liste, 3); 

	if(s < 0)
	{
		// fprintf(stderr, "Read "); 
		// perror(i2cdev); 
		return(-1.0);
	}

	aa = liste[0] + (liste[1] << 8) + (liste[2] << 16);
	fa = (double) aa * ftick() / 61.0;

	printf("F=0x%02x %02x %02x = %d = %G\n", 
		liste[2], liste[1], liste[0], aa, fa);

	return(fa); 
}


/* Measure the analog input. */
int ameas(int fd, int ba)
{
	int s; 
	char liste[20];
	int aa;

	s = ioctl(fd, I2C_SLAVE, ba); 
	if(s == -1) 
	{
		// fprintf(stderr, "Bus address %d ", ba); 
		// perror("I2C_SLAVE");
		return(-1);
	}

	s = i2c_smbus_write_byte(fd, ADCSTART_0); 

	if(s < 0)
	{
		// fprintf(stderr, "smbus_write_byte ADCSTART_0 "); 
		// perror(i2cdev); 
		return(-1);
	}

	delay(50);

	s = read(fd, liste, 2); 

	if(s < 0)
	{
		// fprintf(stderr, "Read "); 
		// perror(i2cdev); 
		return(-1);
	}

	aa = liste[0] + (liste[1] << 8);

	printf("Read bytes %d %d A = %d\n", liste[0], liste[1], aa); 

	return(aa);
}





