
#include <stdio.h>

#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>

#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>

/* Eat SIGHUPs */ 
void sighupcatcher(int sig, siginfo_t *sinfo, void *vpx)
{
	return;
}

/* Catch SIGCHILD signals to clean up after dead child processes. */ 
void sigchldcatcher(int sig, siginfo_t *sinfo, void *vpx)
{
    while( waitpid(-1, 0, WNOHANG) > 0);
}

int G_runflag; 

/* Catch SIGINTs so as to stop gracefully. */ 
void sigintcatcher(int sig, siginfo_t *sinfo, void *vpx)
{
	fprintf(stderr, "%s(%d): Caught SIGINT %d\n", __FILE__, __LINE__, sig);
	G_runflag = 0; 
	
	/* Send SIGINTs to all the children */ 
	kill(0, SIGINT);

	/* The sigchldcatcher will wait and clean them up */ 
	_exit(sig);
}


/* Run the server loop. 
	This may return if the inside function decided to break, 
	in which case it returns a value from that, or 
	there might be some other error, in which case -1 is returned
	and errno is set. 

	The workfunction is called as:

	(*workfunction)(clientsockethandle, exparam); 

	and should perform all the required or desired recv and send
	actions itself.  It should return 0 to keep the server going, 
	and any other value to stop the server. 
*/ 

int TCPserverMulti(unsigned short portno, int verbose, 
	int (*workfunction)(), void *exparam, uid_t runuid) 
{
	struct sigaction sact; 

	int hs;
	struct sockaddr_in addr;
	int cc;

	socklen_t alen;
	struct sockaddr_in cliaddr;
	int csd; 
	pid_t cpid; 

	char cliatxt[100];

	/* Eat SIGHUPs */ 
	memset(&sact, 0, sizeof(struct sigaction) ); 
	sact.sa_sigaction = sighupcatcher;
	sact.sa_flags = SA_SIGINFO | SA_RESTART;
	sigaction(SIGHUP, &sact, 0);

	/* Handle SIGCHLD to clean up after child-processes */ 
	memset(&sact, 0, sizeof(struct sigaction));
	sact.sa_sigaction = sigchldcatcher; 
	sact.sa_flags = SA_SIGINFO | SA_RESTART;
	sigaction(SIGCHLD, &sact, 0); 

	/* Handle SIGINT to stop gracefully */
	G_runflag = 1; 
	memset(&sact, 0, sizeof(struct sigaction));
	sact.sa_sigaction = sigintcatcher; 
	sact.sa_flags = SA_SIGINFO | SA_RESTART;
	sigaction(SIGINT, &sact, 0); 

	/* Unless we are being chatty about what is going on 
		inside our program, Go into "Tommy"-mode (as in Tommy 
		the deaf, dumb, and blind kid from the Who's rock opera). 
		This appears to be what is needed for the 
		process to survive the closing of the 
		outside shell. Eating SIGHUPs don't seem to be enough... */ 
	if(!verbose)
	{
		fclose(stdout);
		fclose(stdin);
		fclose(stderr);
	}

	errno = 0;
	hs = socket(PF_INET, SOCK_STREAM, 0);

	if(verbose)
		fprintf(stderr, "%s(%d): socket() = %d errno = %d (%s)\n", 
			__FILE__, __LINE__, hs, errno, strerror(errno));

	if(hs < 0) return(-1);

	memset( (char *) &addr, 0, sizeof(struct sockaddr_in));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(portno);
	addr.sin_addr.s_addr = INADDR_ANY;

	errno = 0;
	cc = bind(hs, (struct sockaddr *) &addr, sizeof(struct sockaddr_in));

	if(verbose)
		fprintf(stderr, "%s(%d): bind(0x%lx,  %d) = %d errno = %d (%s)\n", 
			__FILE__, __LINE__, 
		(unsigned long) addr.sin_addr.s_addr, 
		ntohs(addr.sin_port), 
		cc, errno, strerror(errno));

	if(cc == -1) return(-1);

	/* Having bound, go down to become the "winpcs" user instead. */ 
	if(runuid) 
	{
		errno = 0;
		cc = setuid(runuid);
		if(verbose)
			fprintf(stderr, "%s(%d): setuid(%lu) = %d errno = %d (%s)\n", 
				__FILE__, __LINE__, (unsigned long) runuid, cc, errno, strerror(errno));
		if(cc == -1) return(-1);
	}

	errno = 0;
	cc = listen(hs, 20);

	if(verbose)
		fprintf(stderr, "%s(%d): listen() = %d errno = %d (%s)\n", 
			__FILE__, __LINE__, cc, errno, strerror(errno));

	if(cc == -1) return(-1);

	/* This is the server's accept loop where it starts talking with the 
		client. We have just a simple loop for a simple job here, which runs 
		forever.

		be set up to run in the background as a daemon, and it may want to
		receive and act on signals such as SIGHUP, SIGINT, and if it produces
		child processes
	*/ 

	while(G_runflag)
	{
		alen = sizeof(struct sockaddr_in);		
		errno = 0;
		csd = accept(hs, (struct sockaddr *) &cliaddr, &alen);

		strcpy(cliatxt, inet_ntoa(cliaddr.sin_addr));

		if(verbose)
			fprintf(stderr, "%s(%d): Server has accepted: %d from %s\n", 
				__FILE__, __LINE__, csd, cliatxt);

		if(csd > 0)
		{
			cpid = fork();

			if(cpid == 0)
			{
				/* perform the actual job, done in a separate process, 
					so that can take whatever time it likes. */
				(*workfunction)(csd, exparam, cliatxt, verbose); 

				close(csd);
				_exit(0);
			}

			/* Parent will also have to close the client-socket */ 
			close(csd);
		}
		else
		{
			if(verbose) 
				fprintf(stderr, "%s(%d): accept() = %d errno = %d (%s)\n", 
					__FILE__, __LINE__, csd, errno, strerror(errno));
		}
	}
	shutdown(hs, SHUT_RDWR);
	close(hs);
	return(0);
}

/* Loop select/recv until we are finished. 
	Returns the length of data received in possibly 
	multiple recv() calls, with some specified time-out value.  
	Useful when we don't want to use length-prefixed data,
	or there is a browser or simple tclit program or some such.
*/ 
int BuffRecv(int clisock, char *buff, int buffmax, long tousec)
{
	struct timeval tv;
	int psstat;
	fd_set rfds;
	int nrd, len; 

	len = 0; 
	for(;;)
	{
		FD_ZERO(&rfds);
		FD_SET(clisock, &rfds);

		tv.tv_sec = 0; 
		tv.tv_usec = 10 * 1000;  /* 10 ms */ 
		/* if this had been the struct timespec of pselect()
			tv.tv_nsec = 10 * 1000000;  */ /* 10 ms */ 

		psstat = select(clisock+1, &rfds, NULL, NULL, &tv);

		/* fprintf(stderr, 
			"%s(%d): Select returned %d R:%d \n", 
			__FILE__, __LINE__, psstat, FD_ISSET(clisock, &rfds) );  */ 

		/* Return 0 for time-out 
			return >0 and FD_ISSET(clisock, &rfds);	
		*/ 

		/* If we timed out, we don't have more stuff to read? */ 
		if(psstat == 0) break; 

		if(FD_ISSET(clisock, &rfds))
		{
			nrd = recv(clisock, buff+len, buffmax-len, 0); 

			/* fprintf(stderr, 
				"%s(%d): Received %d\n", 
				__FILE__, __LINE__, nrd);  */ 

			/* // if(psstat == 0) break;  */ 
			len += nrd; 
		}
	} /* forever loop  */ 
	return(len); 
}



/* Run the server loop. 
	This may return if the inside function decided to break, 
	in which case it returns a value from that, or 
	there might be some other error, in which case -1 is returned
	and errno is set. 

	The workfunction is called as:

	(*workfunction)(clientsockethandle, exparam); 

	and should perform all the required or desired recv and send
	actions itself.  It should return 0 to keep the server going, 
	and any other value to stop the server. 
*/ 

int TCPserverSingle(unsigned short portno, int verbose, 
	int (*workfunction)(), void *exparam) 
{
	struct sigaction sact; 

	int hs;
	struct sockaddr_in addr;
	int cc;

	unsigned int alen;
	struct sockaddr_in cliaddr;
	int csd; 
	int wz;

	/* Eat SIGHUPs */ 
	bzero(&sact, sizeof(struct sigaction) ); 
	sact.sa_sigaction = sighupcatcher;
	sact.sa_flags = SA_SIGINFO | SA_RESTART;
	sigaction(SIGHUP, &sact, 0);

#if(0)
	/* Unless we are being chatty about what is going on 
		inside our program, Go into "Tommy"-mode (as in Tommy 
		the deaf, dumb, and blind kid from the Who's rock opera). 
		This appears to be what is needed for the 
		process to survive the closing of the 
		outside shell. Eating SIGHUPs don't seem to be enough... */ 
	if(!verbose)
	{
		fclose(stdout);
		fclose(stdin);
		fclose(stderr);
	}
#endif

	errno = 0;
	hs = socket(PF_INET, SOCK_STREAM, 0);

	if(verbose)
		fprintf(stderr, "%s(%d): socket() = %d errno = %d (%s)\n", 
			__FILE__, __LINE__, hs, errno, strerror(errno));

	if(hs < 0) return(-1);

	memset( (char *) &addr, 0, sizeof(struct sockaddr_in));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(portno);
	addr.sin_addr.s_addr = INADDR_ANY;

	errno = 0;
	cc = bind(hs, (struct sockaddr *) &addr, sizeof(struct sockaddr_in));

	if(verbose)
		fprintf(stderr, "%s(%d): bind(0x%lx,  %d) = %d errno = %d (%s)\n", 
			__FILE__, __LINE__, 
		(unsigned long) addr.sin_addr.s_addr, 
		ntohs(addr.sin_port), 
		cc, errno, strerror(errno));

	if(cc == -1) return(-1);

	errno = 0;
	cc = listen(hs, 20);

	if(verbose)
		fprintf(stderr, "%s(%d): listen() = %d errno = %d (%s)\n", 
			__FILE__, __LINE__, cc, errno, strerror(errno));

	if(cc == -1) return(-1);

	/* This is the server's accept loop where it starts talking with the 
		client. We have just a simple loop for a simple job here, which runs 
		forever.

		be set up to run in the background as a daemon, and it may want to
		receive and act on signals such as SIGHUP, SIGINT, and if it produces
		child processes
	*/ 

	wz = 0; 

	for(;;)
	{
		alen = sizeof(struct sockaddr_in);		
		errno = 0;
		csd = accept(hs, (struct sockaddr *) &cliaddr, &alen);

		if(verbose)
			fprintf(stderr, "%s(%d): Server has accepted: %d\n", 
				__FILE__, __LINE__, csd);

		if(csd > 0)
		{
			/* Call the function that will want to do the recv and send
				work itself. It may want to do this in several goes. */ 
			wz = (*workfunction)(csd, exparam); 

			/* We're done, so just close the client connection. */ 
			shutdown(csd, SHUT_RDWR);
			close(csd);
		}
		else
		{
			if(verbose) 
				fprintf(stderr, "%s(%d): accept() = %d errno = %d (%s)\n", 
					__FILE__, __LINE__, csd, errno, strerror(errno));
		}

		if(wz != 0) break; 
	}
	shutdown(hs, SHUT_RDWR);
	close(hs);
	return(wz);
}



